@@ -47,7 +47,8 @@ Create TYPE "TransactionErrorCode" AS ENUM (
     'WouldExceedMaxAccountCostLimit',
     'WouldExceedMaxBlockCostLimit',
     'UnsupportedVersion',
-    'InvalidWritableAccount'
+    'InvalidWritableAccount',
+    'WouldExceedMaxAccountDataCostLimit'
 );
 
 CREATE TYPE "TransactionError" AS (
@@ -114,7 +115,7 @@ CREATE TYPE "TransactionMessage" AS (
 );
 
 CREATE TYPE "TransactionMessageAddressTableLookup" AS (
-    account_key: BYTEA[],
+    account_key BYTEA,
     writable_indexes SMALLINT[],
     readonly_indexes SMALLINT[]
 );
@@ -152,6 +153,16 @@ CREATE TABLE transaction (
     CONSTRAINT transaction_pk PRIMARY KEY (slot, signature)
 );
 
+-- The table storing block metadata
+CREATE TABLE block (
+    slot BIGINT PRIMARY KEY,
+    blockhash VARCHAR(44),
+    rewards "Reward"[],
+    block_time BIGINT,
+    block_height BIGINT,
+    updated_on TIMESTAMP NOT NULL
+);
+
 /**
  * The following is for keeping historical data for accounts and is not required for plugin to work.
  */
